package back

import (
	"context"
	"errors"
	"log"
	"os"

	"crawshaw.io/sqlite"
	"crawshaw.io/sqlite/sqlitex"
	"github.com/rickb777/date"
)

type GenString func(length int) (string, error)

type Store struct {
	pool      *sqlitex.Pool
	genString GenString
}

func NewStore(filename string, genString GenString) (*Store, error) {
	var needCreate bool
	if _, err := os.Stat(filename); os.IsNotExist(err) {
		needCreate = true
	}
	// If the file exists, then assume it was created properly.

	pool, err := sqlitex.Open(filename, 0, 10)
	if err != nil {
		return nil, err
	}
	store := &Store{
		pool:      pool,
		genString: genString,
	}
	if needCreate {
		log.Println("creating schema")
		err = store.createSchema()
		if err != nil {
			defer pool.Close()
			return nil, err
		}
	}

	return store, nil
}

func NewMemoryStore(genString GenString) (*Store, error) {
	pool, err := sqlitex.Open("file::memory:?mode=memory&cache=shared", 0, 10)
	if err != nil {
		return nil, err
	}
	store := &Store{
		pool:      pool,
		genString: genString,
	}
	log.Println("creating schema")
	err = store.createSchema()
	if err != nil {
		defer pool.Close()
		return nil, err
	}
	return store, nil
}

func (s *Store) Close() error {
	return s.pool.Close()
}

const schema = `
	CREATE TABLE event (
		id INTEGER PRIMARY KEY,
		alpha_id TEXT NOT NULL,
		admin_code TEXT NOT NULL,
		name TEXT NOT NULL,
		description TEXT NOT NULL,
		earliest_date DATE NOT NULL,
		latest_date DATE NOT NULL,

		UNIQUE (alpha_id)
	);

	CREATE TABLE response (
		id INTEGER PRIMARY KEY,
		alpha_id TEXT NOT NULL,
		event_id INTEGER NOT NULL,

		UNIQUE (alpha_id),
		FOREIGN KEY (event_id) REFERENCES event(id)
	);

	CREATE TABLE response_time (
		response_id INTEGER PRIMARY KEY,
		date DATE NOT NULL,
		time INTEGER NOT NULL,

		CHECK (0 <= time < 24),
		UNIQUE (response_id, date, time)
	);
	
	CREATE TRIGGER enforce_valid_response_dates
	BEFORE INSERT ON response_time
	BEGIN
		SELECT
			CASE
				WHEN NEW.date BETWEEN event.earliest_date AND event.latest_date
					THEN RAISE(ABORT, 'response date is out of range')
			END
		FROM response
		JOIN event ON response.event_id = event.id
		WHERE response.id = NEW.response_id;
	END;`

func (s *Store) createSchema() error {
	conn := s.pool.Get(context.Background())
	defer s.pool.Put(conn)
	return sqlitex.ExecScript(conn, schema)
}

type CreateEventCommand struct {
	Name             string
	Description      string
	Earliest, Latest date.Date
}

type CreateEventResult struct {
	AlphaID, AdminCode string
}

const dbDateLayout = "2006-01-02"

func (s *Store) CreateEvent(ctx context.Context, cmd CreateEventCommand) (result CreateEventResult, err error) {
	const alphaIDLength = 10
	const adminCodeLength = 10

	conn := s.pool.Get(ctx)
	defer s.pool.Put(conn)

	alphaID, err := s.genString(alphaIDLength)
	if err != nil {
		return
	}

	adminCode, err := s.genString(adminCodeLength)
	if err != nil {
		return
	}

	const query = `
		INSERT INTO event(alpha_id, admin_code, name, description, earliest_date, latest_date)
		VALUES (?, ?, ?, ?, ?, ?);`
	err = sqlitex.Exec(conn, query, nil,
		alphaID,
		adminCode,
		cmd.Name,
		cmd.Description,
		cmd.Earliest.Format(dbDateLayout),
		cmd.Latest.Format(dbDateLayout),
	)
	if err != nil {
		return
	}

	result.AdminCode = adminCode
	result.AlphaID = alphaID
	return
}

type GetEventMetadataQuery struct {
	AlphaID string
}

type GetEventMetadataResult struct {
	Name             string
	Description      string
	Earliest, Latest date.Date
}

func (s *Store) GetEventMetadata(ctx context.Context, query GetEventMetadataQuery) (GetEventMetadataResult, error) {
	conn := s.pool.Get(ctx)
	defer s.pool.Put(conn)

	const dbQuery = `
		SELECT name, description, earliest_date, latest_date
		FROM event
		WHERE alpha_id = ?;`
	var result GetEventMetadataResult
	var found bool
	err := sqlitex.Exec(conn, dbQuery,
		func(stmt *sqlite.Stmt) error {
			found = true

			result.Name = stmt.ColumnText(0)
			result.Description = stmt.ColumnText(1)
			earliestDateString := stmt.ColumnText(2)
			latestDateString := stmt.ColumnText(3)

			var err error
			result.Earliest, err = date.Parse(dbDateLayout, earliestDateString)
			if err != nil {
				return err
			}
			result.Latest, err = date.Parse(dbDateLayout, latestDateString)
			return err
		}, query.AlphaID)
	if err != nil {
		return GetEventMetadataResult{}, err
	}
	if !found {
		// TODO return a constant or a specific error type for Not Found
		return GetEventMetadataResult{}, errors.New("not found")
	}
	return result, nil
}

type GetEventResponsesQuery struct {
	AlphaID string
}

type GetEventResponsesResult struct {
	TotalResponses int
}

func (s *Store) GetEventResponses(ctx context.Context, query GetEventResponsesQuery) (GetEventResponsesResult, error) {
	conn := s.pool.Get(ctx)
	defer s.pool.Put(conn)

	var result GetEventResponsesResult
	const dbQuery = `
		SELECT COUNT(*)
		FROM response
		JOIN event ON response.event_id = event.id
		WHERE event.alpha_id = ?;`
	err := sqlitex.Exec(conn, dbQuery,
		func(stmt *sqlite.Stmt) error {
			result.TotalResponses = stmt.ColumnInt(0)
			return nil
		}, query.AlphaID)
	// TODO return an error if the event does not exist?
	return result, err
}

type CreateEventResponseCommand struct {
	EventAlphaID string
	DateHours    map[date.Date]int
}

type CreateEventResponseResult struct {
	ResponseAlphaID string
}

func (s *Store) CreateEventResponse(ctx context.Context, cmd CreateEventResponseCommand) (result CreateEventResponseResult, err error) {
	const responseAlphaIDLength = 10

	conn := s.pool.Get(ctx)
	defer s.pool.Put(conn)

	defer sqlitex.Save(conn)(&err)
	responseAlphaID, err := s.genString(responseAlphaIDLength)
	if err != nil {
		return CreateEventResponseResult{}, err
	}

	// Create response
	const responseQuery = `
		INSERT INTO response(event_id, alpha_id)
			SELECT event.id AS event_id, ? AS alpha_id
			FROM event
			WHERE event.alpha_id = ?;`
	err = sqlitex.Exec(conn, responseQuery, nil, responseAlphaID, cmd.EventAlphaID)
	if err != nil {
		return CreateEventResponseResult{}, err
	}
	// TODO Create response times

	return CreateEventResponseResult{
		ResponseAlphaID: responseAlphaID,
	}, nil
}
